'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var assign = Object.assign;

function addToArray(a, b) {
  var i = a.indexOf(b);
  if (i === -1) {
    a.push(b);
  }
}

function removeFromArray(a, b) {
  var i = a.indexOf(b);
  if (i !== -1) {
    a.splice(i, 1);
  }
}

var _nextId = 0;
function nextId() {
  return _nextId++;
}

var unique = Object.freeze({ equals: function () { return false; } });

function some(x) {
  return x !== null && x !== void 0;
}

var DEBUG_MODE = false;

function setDebugMode(val) {
  DEBUG_MODE = !!val;
}

function isDebug() {
  return DEBUG_MODE;
}

function defaultEquals(a, b) {
  return (
    Object.is(a, b) || (a && typeof a.equals === "function" && a.equals(b))
  );
}

function setEquals(derivable, eq) {
  derivable._equals = eq;
  return derivable;
}

function equals(ctx, a, b) {
  return (ctx._equals || defaultEquals)(a, b);
}

var ATOM = "ATOM";
var DERIVATION = "DERIVATION";
var LENS = "LENS";
var REACTOR = "REACTOR";

function isDerivable(x) {
  return x && (x._type === DERIVATION || x._type === ATOM || x._type === LENS);
}

function isAtom(x) {
  return x && (x._type === ATOM || x._type === LENS);
}

function isDerivation(x) {
  return x && (x._type === DERIVATION || x._type === LENS);
}

function isLens(x) {
  return x && x._type === LENS;
}

var UNKNOWN = 0;
var CHANGED = 1;
var UNCHANGED = 2;
var DISCONNECTED = 3;

var parentsStack = [];
var child = null;

function startCapturingParents(_child, parents) {
  parentsStack.push({ parents: parents, offset: 0, child: _child });
  child = _child;
}
function retrieveParentsFrame() {
  return parentsStack[parentsStack.length - 1];
}
function stopCapturingParents() {
  parentsStack.pop();
  child =
    parentsStack.length === 0
      ? null
      : parentsStack[parentsStack.length - 1].child;
}

function maybeCaptureParent(p) {
  if (child !== null) {
    var frame = parentsStack[parentsStack.length - 1];
    if (frame.parents[frame.offset] === p) {
      // nothing to do, just skip over
      frame.offset++;
    } else {
      // look for this parent elsewhere
      var idx = frame.parents.indexOf(p);
      if (idx === -1) {
        // not seen this parent yet, add it in the correct place
        // and push the one currently there to the end (likely that we'll be
        // getting rid of it)
        // sneaky hack for doing captureDereferences
        if (child !== void 0) {
          addToArray(p._activeChildren, child);
        }
        if (frame.offset === frame.parents.length) {
          frame.parents.push(p);
        } else {
          frame.parents.push(frame.parents[frame.offset]);
          frame.parents[frame.offset] = p;
        }
        frame.offset++;
      } else {
        if (idx > frame.offset) {
          // seen this parent after current point in array, so swap positions
          // with current point's parent
          var tmp = frame.parents[idx];
          frame.parents[idx] = frame.parents[frame.offset];
          frame.parents[frame.offset] = tmp;
          frame.offset++;
        }
        // else seen this parent at previous point and so don't increment offset
      }
    }
  }
}

function captureDereferences(f) {
  var captured = [];
  startCapturingParents(void 0, captured);
  try {
    f();
  } finally {
    stopCapturingParents();
  }
  return captured;
}

function Derivation(deriver, meta) {
  if ( meta === void 0 ) meta = null;

  this._deriver = deriver;
  this._parents = null;
  this._type = DERIVATION;
  this._value = unique;
  this._equals = null;
  this._activeChildren = [];
  this._state = DISCONNECTED;
  this._meta = meta;

  if (isDebug()) {
    this.stack = Error().stack;
  }
}

assign(Derivation.prototype, {
  _clone: function _clone() {
    return setEquals(derive(this._deriver), this._equals);
  },

  _forceEval: function _forceEval() {
    var this$1 = this;

    var newVal = null;
    var newNumParents;

    try {
      if (this._parents === null) {
        this._parents = [];
      }
      startCapturingParents(this, this._parents);
      if (!isDebug()) {
        newVal = this._deriver();
      } else {
        try {
          newVal = this._deriver();
        } catch (e) {
          console.error(this.stack);
          throw e;
        }
      }
      newNumParents = retrieveParentsFrame().offset;
    } finally {
      stopCapturingParents();
    }

    if (!equals(this, newVal, this._value)) {
      this._state = CHANGED;
    } else {
      this._state = UNCHANGED;
    }

    for (var i = newNumParents, len = this._parents.length; i < len; i++) {
      var oldParent = this$1._parents[i];
      detach(oldParent, this$1);
      this$1._parents[i] = null;
    }

    this._parents.length = newNumParents;

    this._value = newVal;
  },

  _update: function _update() {
    var this$1 = this;

    if (this._parents === null) {
      // this._state === DISCONNECTED
      this._forceEval();
      // this._state === CHANGED ?
    } else if (this._state === UNKNOWN) {
      var len = this._parents.length;
      for (var i = 0; i < len; i++) {
        var parent = this$1._parents[i];

        if (parent._state === UNKNOWN) {
          parent._update();
        }

        if (parent._state === CHANGED) {
          this$1._forceEval();
          break;
        }
      }
      if (this._state === UNKNOWN) {
        this._state = UNCHANGED;
      }
    }
  },

  get: function get() {
    maybeCaptureParent(this);
    if (this._activeChildren.length > 0) {
      this._update();
    } else {
      startCapturingParents(void 0, []);
      try {
        this._value = this._deriver();
      } finally {
        stopCapturingParents();
      }
    }
    return this._value;
  }
});

function detach(parent, child) {
  removeFromArray(parent._activeChildren, child);
  if (parent._activeChildren.length === 0 && parent._parents != null) {
    var len = parent._parents.length;
    for (var i = 0; i < len; i++) {
      detach(parent._parents[i], parent);
    }
    parent._parents = null;
    parent._state = DISCONNECTED;
  }
}

function derive(f, meta) {
  if (typeof f !== "function") {
    throw Error("derive requires function");
  }
  return new Derivation(f, meta);
}

function Reactor(parent, react) {
  this._parent = parent;
  this.react = react;
  this._governor = null;
  this._active = false;
  this._reacting = false;
  this._type = REACTOR;

  if (isDebug()) {
    this.stack = Error().stack;
  }
}

assign(Reactor.prototype, {
  start: function start() {
    this._active = true;

    addToArray(this._parent._activeChildren, this);

    this._parent.get();
  },

  _force: function _force(nextValue) {
    try {
      this._reacting = true;
      this.react(nextValue);
    } catch (e) {
      if (isDebug()) {
        console.error(this.stack);
      }
      throw e;
    } finally {
      this._reacting = false;
    }
  },

  force: function force() {
    this._force(this._parent.get());
  },

  stop: function stop() {
    detach(this._parent, this);
    this._active = false;
  },

  _maybeReact: function _maybeReact() {
    if (!this._reacting && this._active) {
      if (this._governor !== null) {
        this._governor._maybeReact();
      }
      // maybe the reactor was stopped by the parent
      if (this._active) {
        var nextValue = this._parent.get();
        if (this._parent._state === CHANGED) {
          this._force(nextValue);
        }
      }
    }
  }
});

function makeReactor(derivable, f, opts) {
  if (typeof f !== "function") {
    throw Error("the first argument to .react must be a function");
  }

  opts = assign(
    {
      once: false,
      skipFirst: false
    },
    opts
  );

  var skipFirst = opts.skipFirst;

  // wrap reactor so f doesn't get a .this context, and to allow
  // stopping after one reaction if desired.
  var reactor = new Reactor(derivable, function (val) {
    if (skipFirst) {
      skipFirst = false;
    } else {
      f(val);
      if (opts.once) {
        reactor.stop();
        controller.stop();
      }
    }
  });

  var assertCondition = function (condition, name) {
    if (isDerivable(condition)) {
      return condition;
    }
    if (typeof condition === "function") {
      return condition;
    }
    if (typeof condition === "undefined") {
      return condition;
    }
    throw Error(
      "react " + name + " condition must be derivable or function, got: " +
        JSON.stringify(condition)
    );
  };

  var getCondition = function (condition, def) { return condition
      ? typeof condition === "function" ? condition(derivable) : condition.get()
      : def; };

  // listen to from condition, starting the reactor controller
  // when appropriate
  var $from = assertCondition(opts.from, "from");
  // listen to when and until conditions, starting and stopping the
  // reactor as appropriate, and stopping this controller when until
  // condition becomes true
  var $until = assertCondition(opts.until, "until");
  var $when = assertCondition(opts.when, "when");

  var $conds = derive(function () {
    return {
      from: getCondition($from, true),
      until: getCondition($until, false),
      when: getCondition($when, true)
    };
  });

  var started = false;

  var controller = new Reactor($conds, function (conds) {
    if (conds.from) {
      started = true;
    }
    if (started) {
      if (conds.until) {
        reactor.stop();
        controller.stop();
      } else if (conds.when) {
        if (!reactor._active) {
          reactor.start();
          reactor.force();
        }
      } else if (reactor._active) {
        reactor.stop();
      }
    }
  });
  controller.start();
  controller.force();

  reactor._governor = controller;
}

/**
 * dereferences a thing if it is dereferencable, otherwise just returns it.
 */

function unpack(thing) {
  if (isDerivable(thing)) {
    return thing.get();
  } else {
    return thing;
  }
}

function deepUnpack(thing) {
  if (isDerivable(thing)) {
    return thing.get();
  } else if (Array.isArray(thing)) {
    return thing.map(deepUnpack);
  } else if (thing.constructor === Object) {
    var result = {};
    var keys = Object.keys(thing);
    for (var i = keys.length; i--; ) {
      var prop = keys[i];
      result[prop] = deepUnpack(thing[prop]);
    }
    return result;
  } else {
    return thing;
  }
}

function struct(arg) {
  if (arg.constructor === Object || Array.isArray(arg)) {
    return derive(function () { return deepUnpack(arg); });
  } else {
    throw new Error("`struct` expects plain Object or Array");
  }
}

var derivablePrototype = {
  derive: function derive$1(f) {
    var this$1 = this;

    if (typeof f !== "function") {
      throw Error("derive requires function");
    }
    return derive(function () { return f(this$1.get()); });
  },

  maybeDerive: function maybeDerive(f) {
    var this$1 = this;

    if (typeof f !== "function") {
      throw Error("maybeDerive requires function");
    }
    return derive(function () {
      var arg = this$1.get();
      return some(arg) ? f(arg) : null;
    });
  },

  orDefault: function orDefault(def) {
    if (!some(def)) {
      throw Error("orDefault requires non-null value");
    }
    return this.derive(function (value) { return (some(value) ? value : def); });
  },

  react: function react(f, opts) {
    makeReactor(this, f, opts);
  },

  maybeReact: function maybeReact(f, opts) {
    var maybeWhen = this.derive(Boolean);
    if (opts && "when" in opts && opts.when !== true) {
      var when = opts.when;
      if (typeof when === "function" || when === false) {
        when = derive(when);
      } else if (!isDerivable(when)) {
        throw new Error("when condition must be bool, function, or derivable");
      }
      maybeWhen = maybeWhen.derive(function (d) { return d && when.get(); });
    }
    makeReactor(this, f, assign({}, opts, { when: maybeWhen }));
  },

  is: function is(other) {
    var this$1 = this;

    return derive(function () { return equals(this$1, this$1.get(), unpack(other)); });
  },

  withEquality: function withEquality(equals$$1) {
    if (equals$$1) {
      if (typeof equals$$1 !== "function") {
        throw new Error("equals must be function");
      }
    } else {
      equals$$1 = null;
    }

    return setEquals(this._clone(), equals$$1);
  }
};

var mutablePrototype = {
  update: function update(f, a, b, c, d) {
    switch (arguments.length) {
      case 0:
        throw Error("update method accepts at least 1 argument");
      case 1:
        return this.set(f(this.get()));
      case 2:
        return this.set(f(this.get(), a));
      case 3:
        return this.set(f(this.get(), a, b));
      case 4:
        return this.set(f(this.get(), a, b, c));
      case 5:
        return this.set(f(this.get(), a, b, c, d));
      default:
        throw Error("update method accepts only 5 arguments");
    }
  }
};

function mark(node, reactors) {
  for (var i = 0, len = node._activeChildren.length; i < len; i++) {
    var child = node._activeChildren[i];
    switch (child._type) {
      case DERIVATION:
      case LENS:
        if (child._state !== UNKNOWN) {
          child._state = UNKNOWN;
          mark(child, reactors);
        }
        break;
      case REACTOR:
        reactors.push(child);
        break;
    }
  }
}

function processReactors(reactors) {
  for (var i = 0, len = reactors.length; i < len; i++) {
    var r = reactors[i];
    if (r._reacting) {
      throw Error(
        "Synchronous cyclical reactions disallowed. " + "Use setImmediate."
      );
    }
    r._maybeReact();
  }
}

var TransactionAbortion = {};

function initiateAbortion() {
  throw TransactionAbortion;
}

function TransactionContext(parent) {
  this.parent = parent;
  this.id2originalValue = {};
  this.modifiedAtoms = [];
}

function maybeTrack(atom) {
  if (currentCtx !== null) {
    if (!(atom._id in currentCtx.id2originalValue)) {
      currentCtx.modifiedAtoms.push(atom);
      currentCtx.id2originalValue[atom._id] = atom._value;
    }
  }
}

var currentCtx = null;

function inTransaction() {
  return currentCtx !== null;
}

function transact(f) {
  beginTransaction();
  try {
    f(initiateAbortion);
  } catch (e) {
    abortTransaction();
    if (e !== TransactionAbortion) {
      throw e;
    }
    return;
  }
  commitTransaction();
}

function atomically(f) {
  if (!inTransaction()) {
    transact(f);
  } else {
    f();
  }
}

function transaction(f) {
  return function () {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    var result;
    transact(function () {
      result = f.apply(void 0, args);
    });
    return result;
  };
}

function atomic(f) {
  return function () {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    var result;
    atomically(function () {
      result = f.apply(void 0, args);
    });
    return result;
  };
}

function beginTransaction() {
  currentCtx = new TransactionContext(currentCtx);
}

function commitTransaction() {
  var ctx = currentCtx;
  currentCtx = ctx.parent;

  if (currentCtx === null) {
    var reactors = [];
    ctx.modifiedAtoms.forEach(function (a) {
      if (equals(a, a._value, ctx.id2originalValue[a._id])) {
        a._state = UNCHANGED;
      } else {
        a._state = CHANGED;
        mark(a, reactors);
      }
    });
    processReactors(reactors);
    ctx.modifiedAtoms.forEach(function (a) {
      a._state = UNCHANGED;
    });
  }
}

function abortTransaction() {
  var ctx = currentCtx;
  currentCtx = ctx.parent;
  ctx.modifiedAtoms.forEach(function (atom) {
    atom._value = ctx.id2originalValue[atom._id];
    atom._state = UNCHANGED;
    mark(atom, []);
  });
}

var _tickerRefCount = 0;

function ticker() {
  if (_tickerRefCount === 0) {
    beginTransaction();
  }
  _tickerRefCount++;
  var done = false;
  return {
    tick: function tick() {
      if (done) { throw new Error("trying to use ticker after release"); }
      commitTransaction();
      beginTransaction();
    },
    reset: function reset() {
      if (done) { throw new Error("trying to use ticker after release"); }
      abortTransaction();
      beginTransaction();
    },
    release: function release() {
      if (done) { throw new Error("ticker already released"); }
      _tickerRefCount--;
      done = true;
      if (_tickerRefCount === 0) {
        commitTransaction();
      }
    }
  };
}

var global$1 = (typeof window === "object"
  ? window
  : typeof global === "object" ? global : {});

var devtoolsHook = global$1.__DERIVABLE_DEVTOOLS_HOOK__;

function Atom(value, meta) {
  if ( meta === void 0 ) meta = null;

  this._id = nextId();
  this._activeChildren = [];
  this._value = value;
  this._state = UNCHANGED;
  this._type = ATOM;
  this._equals = null;
  this._meta = meta;
}

assign(Atom.prototype, {
  _clone: function _clone() {
    return setEquals(atom(this._value), this._equals);
  },

  set: function set(value) {
    maybeTrack(this);

    var oldValue = this._value;
    this._value = value;

    if (!inTransaction()) {
      if (!equals(this, value, oldValue)) {
        try {
          this._state = CHANGED;
          var reactors = [];
          mark(this, reactors);
          processReactors(reactors);
        } finally {
          this._state = UNCHANGED;
        }
      }
    }
  },

  get: function get() {
    if (typeof devtoolsHook === "function") {
      devtoolsHook("captureAtom", this);
    }
    maybeCaptureParent(this);
    return this._value;
  }
});

function atom(value, meta) {
  return new Atom(value, meta);
}

function Lens(descriptor, meta) {
  Derivation.call(this, descriptor.get, meta);
  this._descriptor = descriptor;
  this._type = LENS;
}

assign(Lens.prototype, Derivation.prototype, {
  _clone: function _clone() {
    return setEquals(new Lens(this._descriptor), this._equals);
  },

  set: function set(value) {
    var this$1 = this;

    atomically(function () {
      this$1._descriptor.set(value);
    });
  }
});

function lens(descriptor, meta) {
  return new Lens(descriptor, meta);
}

assign(Derivation.prototype, derivablePrototype);
assign(Lens.prototype, derivablePrototype, mutablePrototype);
assign(Atom.prototype, derivablePrototype, mutablePrototype);

if (global$1.__DERIVABLE_INIT_FLAG__) {
  console.warn(
    "Multiple instances of derivable have been initialized on the same page"
  );
}
global$1.__DERIVABLE_INIT_FLAG__ = true;

exports.atom = atom;
exports.lens = lens;
exports.derive = derive;
exports.setDebugMode = setDebugMode;
exports.isDerivable = isDerivable;
exports.isAtom = isAtom;
exports.isLens = isLens;
exports.isDerivation = isDerivation;
exports.unpack = unpack;
exports.struct = struct;
exports.transact = transact;
exports.transaction = transaction;
exports.ticker = ticker;
exports.atomic = atomic;
exports.atomically = atomically;
exports.__Reactor = Reactor;
exports.__captureDereferences = captureDereferences;
//# sourceMappingURL=derivable.js.map
